@@ -18,14 +18,8 @@ from sklearn.neighbors import KDTree, BallTree
 from sklearn.externals.joblib import Memory
 from sklearn.externals import six
 from warnings import warn
-
-#Try and work around older sklearn api
-try:
-    from sklearn.utils import check_array
-except ImportError:
-    from sklearn.utils import check_arrays
-
-    check_array = check_arrays
+from sklearn.utils import check_array
+from sklearn.externals.joblib.parallel import cpu_count
 
 from scipy.sparse import csgraph
 
@@ -44,7 +38,55 @@ from .dist_metrics import DistanceMetric
 
 from .plots import CondensedTree, SingleLinkageTree, MinimumSpanningTree
 
-FAST_METRICS = KDTree.valid_metrics + BallTree.valid_metrics
+FAST_METRICS = KDTree.valid_metrics + BallTree.valid_metrics + ['cosine', 'arccos']
+
+# Supporting numpy prior to version 1.7 is a little painful ...
+if hasattr(np, 'isclose'):
+    from numpy import isclose
+else:
+    def isclose(a, b, rtol=1.e-5, atol=1.e-8, equal_nan=False):
+
+        def within_tol(x, y, atol, rtol):
+            with np.errstate(invalid='ignore'):
+                result = np.less_equal(abs(x-y), atol + rtol * abs(y))
+            if np.isscalar(a) and np.isscalar(b):
+                result = bool(result)
+            return result
+
+        x = np.array(a, copy=False, subok=True, ndmin=1)
+        y = np.array(b, copy=False, subok=True, ndmin=1)
+
+        # Make sure y is an inexact type to avoid bad behavior on abs(MIN_INT).
+        # This will cause casting of x later. Also, make sure to allow subclasses
+        # (e.g., for numpy.ma).
+        dt = np.core.multiarray.result_type(y, 1.)
+        y = np.array(y, dtype=dt, copy=False, subok=True)
+
+        xfin = np.isfinite(x)
+        yfin = np.isfinite(y)
+        if np.all(xfin) and np.all(yfin):
+            return within_tol(x, y, atol, rtol)
+        else:
+            finite = xfin & yfin
+            cond = np.zeros_like(finite, subok=True)
+            # Because we're using boolean indexing, x & y must be the same shape.
+            # Ideally, we'd just do x, y = broadcast_arrays(x, y). It's in
+            # lib.stride_tricks, though, so we can't import it here.
+            x = x * np.ones_like(cond)
+            y = y * np.ones_like(cond)
+            # Avoid subtraction with infinite/nan values...
+            cond[finite] = within_tol(x[finite], y[finite], atol, rtol)
+            # Check for equality of infinite values...
+            cond[~finite] = (x[~finite] == y[~finite])
+            if equal_nan:
+                # Make NaN == NaN
+                both_nan = np.isnan(x) & np.isnan(y)
+                cond[both_nan] = both_nan[both_nan]
+
+            if np.isscalar(a) and np.isscalar(b):
+                return bool(cond)
+            else:
+                return cond
 
 
 def _tree_to_labels(X, single_linkage_tree, min_cluster_size=10, allow_single_cluster=False):
@@ -62,12 +104,9 @@ def _tree_to_labels(X, single_linkage_tree, min_cluster_size=10, allow_single_cl
 def _hdbscan_generic(X, min_samples=5, alpha=1.0,
                      metric='minkowski', p=2, leaf_size=None, gen_min_span_tree=False, **kwargs):
     if metric == 'minkowski':
-        if p is None:
-            raise TypeError('Minkowski metric given but no p value supplied!')
-        if p < 0:
-            raise ValueError('Minkowski metric with negative p value is not defined!')
-
         distance_matrix = pairwise_distances(X, metric=metric, p=p)
+    elif metric == 'arccos':
+        distance_matrix = pairwise_distances(X, metric='cosine', **kwargs)
     else:
         distance_matrix = pairwise_distances(X, metric=metric, **kwargs)
 
@@ -87,7 +126,7 @@ def _hdbscan_generic(X, min_samples=5, alpha=1.0,
     if gen_min_span_tree:
         result_min_span_tree = min_spanning_tree.copy()
         for index, row in enumerate(result_min_span_tree[1:], 1):
-            candidates = np.where(np.isclose(mutual_reachability_[row[1]], row[2]))[0]
+            candidates = np.where(isclose(mutual_reachability_[int(row[1])], row[2]))[0]
             candidates = np.intersect1d(candidates, min_spanning_tree[:index, :2].astype(int))
             candidates = candidates[candidates != row[1]]
             assert (len(candidates) > 0)
@@ -107,12 +146,6 @@ def _hdbscan_sparse_distance_matrix(X, min_samples=5, alpha=1.0,
                                     metric='minkowski', p=2, leaf_size=40,
                                     gen_min_span_tree=False, **kwargs):
 
-    if metric == 'minkowski':
-        if p is None:
-            raise TypeError('Minkowski metric given but no p value supplied!')
-        if p < 0:
-            raise ValueError('Minkowski metric with negative p value is not defined!')
-
     assert(issparse(X))
 
     lil_matrix = X.tolil()
@@ -145,21 +178,11 @@ def _hdbscan_sparse_distance_matrix(X, min_samples=5, alpha=1.0,
 
 def _hdbscan_prims_kdtree(X, min_samples=5, alpha=1.0,
                           metric='minkowski', p=2, leaf_size=40, gen_min_span_tree=False, **kwargs):
-    if metric == 'minkowski':
-        if p is None:
-            raise TypeError('Minkowski metric given but no p value supplied!')
-        if p < 0:
-            raise ValueError('Minkowski metric with negative p value is not defined!')
-    elif p is None:
-        p = 2  # Unused, but needs to be integer; assume euclidean
 
     # The Cython routines used require contiguous arrays
     if not X.flags['C_CONTIGUOUS']:
         X = np.array(X, dtype=np.double, order='C')
 
-    size = X.shape[0]
-    min_samples = min(size - 1, min_samples)
-
     tree = KDTree(X, metric=metric, leaf_size=leaf_size, **kwargs)
 
     # TO DO: Deal with p for minkowski appropriately
@@ -183,21 +206,11 @@ def _hdbscan_prims_kdtree(X, min_samples=5, alpha=1.0,
 
 def _hdbscan_prims_balltree(X, min_samples=5, alpha=1.0,
                             metric='minkowski', p=2, leaf_size=40, gen_min_span_tree=False, **kwargs):
-    if metric == 'minkowski':
-        if p is None:
-            raise TypeError('Minkowski metric given but no p value supplied!')
-        if p < 0:
-            raise ValueError('Minkowski metric with negative p value is not defined!')
-    elif p is None:
-        p = 2  # Unused, but needs to be integer; assume euclidean
 
     # The Cython routines used require contiguous arrays
     if not X.flags['C_CONTIGUOUS']:
         X = np.array(X, dtype=np.double, order='C')
 
-    size = X.shape[0]
-    min_samples = min(size - 1, min_samples)
-
     tree = BallTree(X, metric=metric, leaf_size=leaf_size, **kwargs)
 
     dist_metric = DistanceMetric.get_metric(metric, **kwargs)
@@ -222,27 +235,20 @@ def _hdbscan_boruvka_kdtree(X, min_samples=5, alpha=1.0,
                             approx_min_span_tree=True,
                             gen_min_span_tree=False,
                             core_dist_n_jobs=4, **kwargs):
-    if metric == 'minkowski':
-        if p is None:
-            raise TypeError('Minkowski metric given but no p value supplied!')
-        if p < 0:
-            raise ValueError('Minkowski metric with negative p value is not defined!')
 
     if leaf_size < 3:
         leaf_size = 3
 
     if core_dist_n_jobs < 1:
-        raise ValueError('Parallel core distance computation requires 1 or more jobs!')
-
-    size = X.shape[0]
-    min_samples = min(size - 1, min_samples)
+       core_dist_n_jobs = max(cpu_count() + 1 + core_dist_n_jobs, 1)
 
     tree = KDTree(X, metric=metric, leaf_size=leaf_size, **kwargs)
     alg = KDTreeBoruvkaAlgorithm(tree, min_samples, metric=metric, leaf_size=leaf_size // 3,
                                  approx_min_span_tree=approx_min_span_tree, n_jobs=core_dist_n_jobs, **kwargs)
     min_spanning_tree = alg.spanning_tree()
     # Sort edges of the min_spanning_tree by weight
-    min_spanning_tree = min_spanning_tree[np.argsort(min_spanning_tree.T[2]), :]
+    row_order = np.argsort(min_spanning_tree.T[2])
+    min_spanning_tree = min_spanning_tree[row_order, :]
     # Convert edge list into standard hierarchical clustering format
     single_linkage_tree = label(min_spanning_tree)
 
@@ -257,20 +263,12 @@ def _hdbscan_boruvka_balltree(X, min_samples=5, alpha=1.0,
                               approx_min_span_tree=True,
                               gen_min_span_tree=False,
                               core_dist_n_jobs=4, **kwargs):
-    if metric == 'minkowski':
-        if p is None:
-            raise TypeError('Minkowski metric given but no p value supplied!')
-        if p < 0:
-            raise ValueError('Minkowski metric with negative p value is not defined!')
 
     if leaf_size < 3:
         leaf_size = 3
 
     if core_dist_n_jobs < 1:
-        raise ValueError('Parallel core distance computation requires 1 or more jobs!')
-
-    size = X.shape[0]
-    min_samples = min(size - 1, min_samples)
+        core_dist_n_jobs = max(cpu_count() + 1 + core_dist_n_jobs, 1)
 
     tree = BallTree(X, metric=metric, leaf_size=leaf_size, **kwargs)
     alg = BallTreeBoruvkaAlgorithm(tree, min_samples, metric=metric, leaf_size=leaf_size // 3,
@@ -365,7 +363,8 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
 
     core_dist_n_jobs : int, optional
         Number of parallel jobs to run in core distance computations (if
-        supported by the specific algorithm).
+        supported by the specific algorithm). For ``core_dist_n_jobs``
+        below -1, (n_cpus + 1 + core_dist_n_jobs) are used.
         (default 4)
 
 
@@ -423,18 +422,29 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
     if min_samples <= 0 or min_cluster_size <= 0:
         raise ValueError('Min samples and Min cluster size must be positive integers')
 
-    if alpha <= 0.0 or type(alpha) is int:
-        raise ValueError('Alpha must be a positive value greater than 0!')
+    if type(alpha) is not float or alpha <= 0.0:
+        raise ValueError('Alpha must be a positive float value greater than 0!')
 
     if leaf_size < 1:
         raise ValueError('Leaf size must be greater than 0!')
 
+    if metric == 'minkowski':
+        if p is None:
+            raise TypeError('Minkowski metric given but no p value supplied!')
+        if p < 0:
+            raise ValueError('Minkowski metric with negative p value is not defined!')
+
     # Checks input and converts to an nd-array where possible
     X = check_array(X, accept_sparse='csr')
     # Python 2 and 3 compliant string_type checking
     if isinstance(memory, six.string_types):
         memory = Memory(cachedir=memory, verbose=0)
 
+    size = X.shape[0]
+    min_samples = min(size - 1, min_samples)
+    if min_samples == 0:
+        min_samples = 1
+
     if algorithm != 'best':
         if algorithm == 'generic':
             (single_linkage_tree, 
@@ -508,7 +518,7 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
             if X.shape[1] > 60:
                 (single_linkage_tree,
                  result_min_span_tree) = \
-                    memory.cache(_hdbscan_prims_kdtree)(X, min_samples, alpha,
+                    memory.cache(_hdbscan_prims_balltree)(X, min_samples, alpha,
                                                         metric, p, leaf_size,
                                                         gen_min_span_tree, **kwargs)
             else:
@@ -745,8 +755,12 @@ class HDBSCAN (BaseEstimator, ClusterMixin):
         if self._outlier_scores is not None:
             return self._outlier_scores
         else:
-            self._outlier_scores = outlier_scores(self._condensed_tree)
-            return self._outlier_scores
+            if self._condensed_tree is not None:
+                self._outlier_scores = outlier_scores(self._condensed_tree)
+                return self._outlier_scores
+            else:
+                warn('No condensed tree was generated; try running fit first.')
+                return None
 
     @property
     def condensed_tree_(self):
@@ -775,7 +789,7 @@ class HDBSCAN (BaseEstimator, ClusterMixin):
                     'No minimum spanning tree will be provided without raw data.')
                 return None
         else:
-            warn('No minimum spanning tree was generated. \n'
-                 'This may be due to optimized algorithm variations that skip\n'
-                 'explicit generation of the spanning tree.')
+            warn('No minimum spanning tree was generated.'
+                'This may be due to optimized algorithm variations that skip'
+                'explicit generation of the spanning tree.')
             return None
